﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.Objects;
using System.Data.Services;
using System.Linq;
using DevExpress.ExpressApp.Security;
using DevExpress.ExpressApp.Utils;
using DevExpress.Persistent.Base;
using DevExpress.Persistent.Base.Security;

namespace DXSK8.Service {
    public partial class SK8DataContainer {
        const int NotificationQuantityThreshold = 5;

        void CheckInventoriesForNotification(IEnumerable<Inventory> inventories) {
            var notifier = new PushNotifications();
            foreach (Inventory inventory in inventories) {
                if (inventory.Quantity < NotificationQuantityThreshold) {
                    NotifyForInventory(inventory, notifier);
                }
            }
        }

        public void NotifyForInventory(Inventory inventory, PushNotifications notifier) {
            foreach (NotificationChannel channel in NotificationChannels) {
                string notificationContent = GetInventoryNotificationXml(inventory);
                notifier.Notify(channel.Url, notificationContent);
            }
        }

        string GetInventoryNotificationXml(Inventory inventory) {
            string text;
            if (inventory.Quantity == 0) {
                text = inventory.Product.Name + ". Item is out of stock!";
            }
            else if (inventory.Quantity == 1) {
                text = inventory.Product.Name + ". Only one item in stock!";
            }
            else {
                text = inventory.Product.Name + ". Only " + inventory.Quantity + " items in stock!";
            }
            string imageSrc = "http://demos.devexpress.com/DXTREME/DXSK8_static/Images/Types/" +
                              inventory.Product.Type.Name.Replace(" ", "").Replace("&", "") + ".png";
            string launchString = string.Format("{0};{1};{2};{3};{4}", inventory.Product.Brand.Name,
                                                inventory.Store.Address.Address1, inventory.Product.Name,
                                                inventory.Quantity, imageSrc);
            return string.Format(@"<?xml version='1.0' encoding='utf-8'?>
                    <toast launch='{0}'>
                        <visual lang='en-US'>
                            <binding template='ToastImageAndText01'>
                                <image id='1' src='{1}' />
                                <text id='1'>{2}</text>
                            </binding>
                        </visual>
                    </toast>
                ", launchString, imageSrc, text);
        }

        void UpdateOrderState() {
            List<OrderItem> changerOrderItems = ObjectStateManager
                .GetObjectStateEntries(EntityState.Modified)
                .Select(entry => entry.Entity)
                .OfType<OrderItem>().ToList();
            changerOrderItems.ForEach(orderItem => {
                if (orderItem.State == 1) {
                    int completedCount = orderItem.Order.OrderItems.Count(item => item.State == 1);
                    orderItem.Order.State =
                        (int)
                        (completedCount == orderItem.Order.OrderItems.Count
                             ? OrderState.Completed
                             : OrderState.Inprogress);
                    UpdateInventory(orderItem.Product.Inventories.First(x => x.Store.Id == 1));
                }
            });
        }

        static void UpdateInventory(Inventory inventory) {
            if (inventory.Quantity == 0) {
                throw new Exception(inventory.Product.Name + " is out of stock");
            }
            inventory.Quantity = inventory.Quantity - 1;
            if (inventory.Quantity == 0) {
                List<Order> orders = inventory.Product.OrderItems
                                              .Where(
                                                  x =>
                                                  (x.Order.State != ((int) OrderState.Completed) && x.State != 1))
                                              .Select(x => x.Order)
                                              .Distinct().ToList();

                orders.ForEach(
                    x => { x.State = ((int) OrderState.OutOfStock); }
                    );
            }
        }

        public override int SaveChanges(SaveOptions options) {
            UpdateOrderState();
            IEnumerable<Inventory> inventories = GetInventories();

            int result = base.SaveChanges(options);

            CheckInventoriesForNotification(inventories);
            return result;
        }

        IEnumerable<Inventory> GetInventories() {
            return ObjectStateManager
                .GetObjectStateEntries(EntityState.Modified)
                .Select(entry => entry.Entity)
                .OfType<Inventory>().ToList();
        }
    }

    public enum OrderState {
        Draft = 0,
        New = 1,
        Inprogress = 2,
        Completed = 3,
        OutOfStock = 4
    }

    [ImageName("BO_User"), DefaultProperty("UserName")]
    public partial class Employee : ISecurityUser,  IAuthenticationStandardUser,
                                    IOperationPermissionProvider {

        #region IOperationPermissionProvider
        IEnumerable<IOperationPermissionProvider> IOperationPermissionProvider.GetChildren() {
            if (!Roles.IsLoaded) {
                Roles.Load();
            }
            return new EnumerableConverter<IOperationPermissionProvider, Role>(Roles);
        }

        IEnumerable<IOperationPermission> IOperationPermissionProvider.GetPermissions() {
            return new IOperationPermission[0];
        }
        #endregion
        #region ISecurityUser
        Boolean ISecurityUser.IsActive {
            get { return IsActive.HasValue && IsActive.Value; }
        }

        String ISecurityUser.UserName {
            get { return UserName; }
        }
        #endregion
        
        #region IAuthenticationStandardUser
        Boolean IAuthenticationStandardUser.ComparePassword(String password) {
            var passwordCryptographer = new PasswordCryptographer();
            return passwordCryptographer.AreEqual(StoredPassword, password);
        }

        public void SetPassword(String password) {
            var passwordCryptographer = new PasswordCryptographer();
            StoredPassword = passwordCryptographer.GenerateSaltedPassword(password);
        }

        Boolean IAuthenticationStandardUser.ChangePasswordOnFirstLogon {
            get { return ChangePasswordOnFirstLogon.HasValue && ChangePasswordOnFirstLogon.Value; }
            set { ChangePasswordOnFirstLogon = value; }
        }

        String IAuthenticationStandardUser.UserName {
            get { return UserName; }
        }
        #endregion
       
    }
    [ImageName("BO_Role")]
    public partial class Role :  IOperationPermissionProvider {
        #region IOperationPermissionProvider

        IEnumerable<IOperationPermissionProvider> IOperationPermissionProvider.GetChildren() {
            return new IOperationPermissionProvider[0];
        }

        IEnumerable<IOperationPermission> IOperationPermissionProvider.GetPermissions() {
            var result = new List<IOperationPermission>();
            foreach (TypePermissionObject persistentPermission in TypePermissions) {
                result.AddRange(persistentPermission.GetPermissions());
            }
            if (IsAdministrative) {
                result.Add(new IsAdministratorPermission());
            }
            if (CanEditModel) {
                result.Add(new ModelOperationPermission());
            }
            return result;
        }
        #endregion

    }
    public partial class TypePermissionObject {
        Type targetType;

        public Type TargetType {
            get {
                if ((targetType == null) && !String.IsNullOrWhiteSpace(TargetTypeFullName)) {
                    targetType = ReflectionHelper.FindType(TargetTypeFullName);
                }
                return targetType;
            }
            set {
                targetType = value;
                TargetTypeFullName = targetType != null ? targetType.FullName : "";
            }
        }

        public IEnumerable<IOperationPermission> GetPermissions() {
            var result = new List<IOperationPermission>();
            if (TargetType != null) {
                if (AllowRead) {
                    result.Add(new TypeOperationPermission(TargetType, SecurityOperations.Read));
                }
                if (AllowWrite) {
                    result.Add(new TypeOperationPermission(TargetType, SecurityOperations.Write));
                }
                if (AllowCreate) {
                    result.Add(new TypeOperationPermission(TargetType, SecurityOperations.Create));
                }
                if (AllowDelete) {
                    result.Add(new TypeOperationPermission(TargetType, SecurityOperations.Delete));
                }
                if (AllowNavigate) {
                    result.Add(new TypeOperationPermission(TargetType, SecurityOperations.Navigate));
                }
            }
            return result;
        }
    }

    public partial class Product : IImageSource {
        #region IImageSource Members
        byte[] IImageSource.Image {
            get { return Image.Image; }
        }
        #endregion
    }

    [IgnoreProperties("Logo")]
    public partial class Brand : IImageSource {
        byte[] IImageSource.Image {
            get { return Logo.Image; }
        }
    }

    [IgnoreProperties("Avatar")]
    public partial class Login : IImageSource {
        #region IImageSource Members
        public byte[] Image {
            get { return Avatar.Image; }
        }
        #endregion
    }

    public partial class ImageSource {
        public void SetImage(byte[] image) {
            Image = image;
        }
    }
}